import zmq
from .. import Problem
from .. import Response
from .. UQOExceptions import *


class Connection:
    """ This class contains methods for the communication handling with the server, e.g. assembling the request message
    with all relevant data, creating a connection and sending requests to the server. A Connection object contains
    the configuration data of the user (url, auth_method and credentials) and additional data referring to the task
    that will be sent to the server. The configuration data is set by initialization as it is passed in the
    create_connection() method in the Config class.


    Attributes
    ----------
    url
        ip+port to which the connection will be established (endpoint)
    auth_method
        authentication method
    credentials
        personal token of the user
    preferred_solver
        Specifies which solver should be used. Available solvers can be displayed by calling the method
        get_available_dwave_solvers().
    preferred_platform
        Specifies which platform should be used. Platforms can be "dwave", "qbsolv" or "uq".
    task
        Specifies the type of task the server should execute.
        Tasks can be "solve", "ping", "dwave_info", "uq_info", "show_quota" or "util".
    solver
        Specifies which solver should be used.
    context
        ZeroMQ-Context

    Methods
    -------
    ping()
        Send a ping message for testing the connection to the server
    find_chimera_embedding(problem)
        Return a chimera embedding for a given problem
    find_pegasus_embedding(problem)
        Return a pegasus embedding for a given problem
    solve_qubo(problem)
        Solve a QUBO problem either with QBsolv or a DWave-Solver
    solve_ising(problem)
        Solve an Ising problem either with QBsolv or a DWave-Solver
    get_available_dwave_solvers()
        Return a list of available solvers from DWave
    get_available_platforms()
        Return a list of available platforms
    get_authentication_message()
        Returns a dictionary that contains the authentication method and the credentials of the user
    get_task_details_message()
        Returns a dictionary that contains information referring to the task
    set_preferred_solver(), set_preferred_platform(), set_task()
        Setter methods for the attributes preferred_solver, preferred_platform and task
    available_tasks()
    send_message(message)
        Create a request socket and connect to the endpoint specified in url. Send the message and return the reply.
    to_json()
    check_errors()
        Check if the message from the server contains an authentication or backend exception
    show_quota()
        Print the time a user has left for computation on a d-wave platform.
    """

    def __init__(self, url, auth_method, credentials):
        """Initialize the connection object. Fill the config data (url, auth_method and credentials) by using the
        passed arguments.

        Parameters
        ----------
        url
            ip+port to which the connection will be established (endpoint)
        auth_method
            authentication method
        credentials
            personal token of the user
        """
        self.url = url
        self.credentials = credentials
        self.auth_method = auth_method
        self.preferred_solver = None
        self.preferred_platform = None
        self.task = None
        self.solver = None
        self.context = zmq.Context().instance()

    # ----------------------- PING MESSAGE ----------------------- #
    def ping(self):
        """Send a ping message to the server. Server should answer with message {"status": "success", "type": "pong"}.

        Returns
        -------
        answer["type"]
            The "pong" from the server
        """
        ping_message = {
            "task": "ping",
            "authentication": self.get_authentication_message()
        }
        answer = self.send_message(ping_message)
        return answer["type"]

    # ----------------------- FIND EMBEDDING MESSAGE ----------------------- #

    def find_chimera_embedding(self, problem):
        """Send a request to the server to find a chimera embedding for a given problem.

        Parameters
        ----------
        problem
            A QUBO or Ising representation of a problem

        Returns
        -------
        embedding
            The Chimera embedding the server sent back
        """

        # Check if problem has a valid format (QUBO or Ising).
        if isinstance(problem, Problem.Qubo) or isinstance(problem, Problem.Ising):
            problem_dict = problem.to_json()
        else:
            raise Exception

        # Create the message that will be sent to the server
        find_embedding_message = {
            "authentication": self.get_authentication_message(),
            "task": "util",
            "task_details": {
                "platform": "dwave",
                "params": {
                    "pref_solver": problem.solver,
                },
                "type": "find_chimera_embedding",
                "problem": problem_dict,
            }
        }

        # Send message to server and save the response in answer
        answer = self.send_message(find_embedding_message)

        # Extract the embedding from the response.
        embedding_stringed = answer["solver_details"]["embedding"]
        embedding = {}
        for key in embedding_stringed:
            embedding[int(key)] = embedding_stringed[key]

        return embedding

    def find_pegasus_embedding(self, problem):
        """Sends a request to the server to find a pegasus embedding for a given problem.

        Parameters
        ----------
        problem
            A QUBO or Ising representation of a problem

        Returns
        -------
        embedding
            Pegasus embedding
        """
        if isinstance(problem, Problem.Qubo) or isinstance(problem, Problem.Ising):
            problem_dict = problem.to_json()
        else:
            raise Exception

        find_embedding_message = {
            "authentication": self.get_authentication_message(),
            "task": "util",
            "task_details": {
                "platform": "dwave",
                "params": {
                    "pref_solver": problem.solver,
                },
                "type": "find_pegasus_embedding",
                "problem": problem_dict,
            }
        }

        answer = self.send_message(find_embedding_message)
        embedding_stringed = answer["solver_details"]["embedding"]

        embedding = {}
        for key in embedding_stringed:
            embedding[int(key)] = embedding_stringed[key]

        return embedding

    # ----------------------- SOLVE QUBOS ----------------------- #

    def solve_qubo(self, problem):
        """Solve a QUBO problem with either QBsolv or a DWave-Solver.

        Returns
        -------
        answer
            The response of type QBSolveResponse or DWaveResponse with the answer from the solver
        """

        # check if problem is in valid QUBO form
        if not isinstance(problem, Problem.Qubo):
            raise NotAQuboException
        else:

            solve_message = {
                "authentication": self.get_authentication_message(),
                "task_details": self.get_task_details_message(problem),
                "task": "solve" if self.task is None else self.task,
            }

            answer = self.send_message(solve_message)

            if answer["status"] == "success":
                if answer["solver"] == "QBsolvSolver":
                    answer = Response.QBSolveResponse(answer["solver_details"]["answer"])
                elif answer["solver"] == "DWaveSolver":
                    answer = Response.DWaveResponse(answer["solver_details"]["answer"])
                return answer
            else:
                self.check_errors(answer)
                print(answer["status"])

                if answer["type"] == "MissingTask":
                    print("Task Fehlt!")
                elif answer["type"] == "InvalidTask":
                    print("UngültigerTask")
                else:
                    raise QBSolveException(answer["message"])

    # ----------------------- SOLVE ISING ----------------------- #

    def solve_ising(self, problem):
        """Solve an Ising problem with either QBsolv or a DWave-Solver.

        Returns
        -------
        answer
            The response of type QBSolveResponse or DWaveResponse with the answer from the solver
        """

        # check if problem is in valid Ising form
        if not isinstance(problem, Problem.Ising):
            raise NotAQuboException
        else:

            solve_message = {
                "authentication": self.get_authentication_message(),
                "task_details": self.get_task_details_message(problem),
                "task": "solve" if self.task is None else self.task,
            }

            answer = self.send_message(solve_message)

            if answer["status"] == "success":
                if answer["solver"] == "QBsolvSolver":
                    answer = Response.QBSolveResponse(answer["solver_details"]["answer"])
                    return answer
                elif answer["solver"] == "DWaveSolver":
                    answer = Response.DWaveResponse(answer["solver_details"]["answer"])

                    return answer
            else:
                print(answer["status"])
                raise QBSolveException(answer["message"])

    # ----------------------- GET DWAVE SOLVERS ----------------------- #

    def get_available_dwave_solvers(self):
        """Return a list of all currently available solvers from DWave. """
        solve_message = {}
        solve_message["authentication"] = self.get_authentication_message()
        solve_message["task"] = "dwave_info"
        solve_message["task_details"] = {"platform": "dwave", "type": "available_solvers"}

        answer = self.send_message(solve_message)
        return list(map(lambda x: str(x), answer["solver_details"]["details"]))

    # ----------------------- GET AVAILABLE PLATFORMS ----------------------- #

    def get_available_platforms(self):
        """Return a list of all currently available platforms. """
        message = {}
        message["authentication"] = self.get_authentication_message()
        message["task"] = "uq_info"
        message["task_details"] = {"type": "available_platforms"}

        answer = self.send_message(message)
        return list(map(lambda x: str(x), answer["details"]))

    # ----------------------------------------  HELPER FUNCTIONS ---------------------------------------- #

    def get_authentication_message(self):
        """Return a dictionary that contains the authentication method and the credentials of the user. """
        return {
            "method": self.auth_method,
            "credentials": self.credentials
        }

    def get_task_details_message(self, problem):
        """Return a dictionary that contains information referring to the task. """
        params = {
            "uq_params": problem.uq_params,
            "solver_params": problem.solver_params
        }

        # if a preferred solver is specified
        if self.preferred_solver is not None:
            params["pref_solver"] = self.preferred_solver

        type = "qubo" if isinstance(problem, Problem.Qubo) else "ising"
        task_details_message = {
            "type": type,
            "task": "solve" if self.task is None else self.task,
            "platform": problem.platform,
            "value": problem.to_json(),
            "params": params
        }
        if self.preferred_platform is not None:
            task_details_message["pref_platform"] = self.preferred_platform

        if problem.embedding is not None:
            task_details_message["embedding"] = problem.embedding

        return task_details_message

    # ----------- setter methods for preferred_solver, preferred_platform and task ----------- #
    def set_preferred_solver(self, preferred_solver):
        self.preferred_solver = preferred_solver

    def set_preferred_platform(self, preferred_platform):
        self.preferred_platform = preferred_platform

    def set_task(self, task):
        self.task = task

    def available_tasks(self):
        return ["solve"]

    def send_message(self, message):
        """Create a request socket and connect to the endpoint specified in url. Send the message and wait for a response message.

        Parameters
        ----------
        message
            The message containing information about the task the server should execute and about authentication data of
            the user.

        Returns
        -------
        answer
            Reply from the server
        """

        socket = self.context.socket(zmq.REQ)  # establish socket
        socket.connect("tcp://" + self.url)

        json_message = self.to_json(message)  # convert message into json object
        socket.send_json(json_message)  # send message

        answer = socket.recv_json()  # wait for response
        self.check_errors(answer)
        return answer

    def to_json(self, message):
        json_message = dict()
        for key in message:
            if type(message[key]) == dict:
                jsonfyed_message = self.to_json(message[key])
                json_message[key] = jsonfyed_message
            elif type(key) == tuple:
                json_message[str(key)] = message[key]
            else:
                json_message[key] = message[key]
        return json_message

    def check_errors(self, answer):
        """Check if the message from the server contains an authentication or backend exception. If there occurs an
        exception, raise an appropriate Exception. For further details about a specific Exception see the collection of
        all exceptions in UQOExceptions.py.

        Parameters
        ----------
        answer
            Response message from the server
        """

        # Check the status of the response. The "type" entry of the response specifies which exception occurred.
        if answer["status"] == "error":
            error_type = answer["type"]

            # Auth-Errors
            if error_type == "InvalidTask":
                raise InvalidTaskException(answer["error_details"])
            elif error_type == "MissingAuthenticationMethod":
                raise AuthMessageMissingException()
            elif error_type == "InvalidAuthenticationMethod":
                raise InvalidAuthMethodException(answer["error_details"])
            elif error_type == "InvalidCredentials":
                raise InvalidAuthCredentials()
            elif error_type == "MissingAuthenticationCredentials":
                raise MalformedAuthMessageException()
            elif error_type == "MissingTask":
                raise MissingTaskException()
            elif error_type == "generic_auth_error":
                raise GenericAuthException()
            elif error_type == "fast_retry_exception":
                raise FastRetryException(answer["error_details"])
            elif error_type == "auth_admin_failed":
                raise AuthAdminFailedException()
            elif error_type == "InsufficientQuota":
                raise InsufficientQuotaException()

            # Backend-Errors

            elif error_type == "generic_backend_error":
                raise GenericBackendException

            elif error_type == "solver_error":
                raise SolverException(answer["error_details"])
            elif error_type == "MissingTask":
                raise MissingTask()
            elif error_type == "InvalidTask":
                raise InvalidTask(answer["error_details"])
            elif error_type == "InvalidSolver":
                raise InvalidSolverException(answer["error_details"])
            elif error_type == "MissingPlatform":
                raise MissingPlatformException(answer["error_details"])

    def show_quota(self):
        """Print the remaining quota (the time you can spend on a DWave platform in microseconds). """

        show_coins_message = {
            "authentication": self.get_authentication_message(),
            "task": "show_quota",
        }

        # send message to server
        answer = self.send_message(show_coins_message)
        print(answer["quota"])




def display_qubo(qubo_dict):
    """Save a QUBO in a well readable format to a txt file.

    Parameters
    ----------
    qubo_dict
        Dictionary that represents the QUBO
    """

    print(qubo_dict)
    max_1 = max(list(map(lambda x: x[0], qubo_dict.keys())))
    max_2 = max(list(map(lambda x: x[1], qubo_dict.keys())))

    amount_of_vars = max_1 if max_1 >= max_2 else max_2

    max_val = -100000000

    for i in range(0, amount_of_vars+1):
        for j in range(0, amount_of_vars + 1):
            if (i, j) in qubo_dict and abs(qubo_dict[(i, j)]) >= max_val:
                max_val = qubo_dict[(i, j)]

    space_per_number = len(str(max_val))
    empty_space = 3

    with open("qubo.txt", "w") as file:
        for i in range(0, amount_of_vars + 1):
            line = ""
            for j in range(0, amount_of_vars + 1):
                if (i, j) in qubo_dict:
                    str_val = str(qubo_dict[(i, j)])
                    line += str_val + " " * (space_per_number - len(str_val) + empty_space)
                else:
                    line += "0" + " " * (space_per_number - 1 + empty_space)

            file.write(line + "\n")



import networkx as nx
import matplotlib.pyplot as plt
from . import util


def create_qubo():
    """Convert a problem instance (graph) into an equivalent QUBO representation.

    Returns
    -------
    QUBO
        Dictionary that represents the QUBO
    """

    # Creating the problem instance
    problem_instance = nx.Graph()
    problem_instance.add_edges_from([(0, 1), (0, 2), (1, 3), (2, 3)])

    # If you want to see the problem instance, uncomment the following lines
    # Note: networkx sometimes draws the graph kinda strange. just execute the drawing multiple times, to get different
    #       equivalent representations of the problem instance

    # The "names" of the nodes in the graph
    nodes = [0, 1, 2, 3]
    # Think of 0 = red, 1 = green, 2 = blue
    colors = [0, 1, 2]

    amount_of_colors = len(colors)

    QUBO = {}

    # Fill the QUBO
    for n_index, node in enumerate(nodes):
        for c_index, color in enumerate(colors):
            # Reward for assigning a color to a node
            QUBO[(amount_of_colors * n_index + c_index, amount_of_colors * n_index + c_index)] = -1
            # Penalty for assigning multiple colors to a node
            for color in range(1, amount_of_colors - (amount_of_colors * n_index + c_index) % amount_of_colors):
                QUBO[(amount_of_colors * n_index + c_index, amount_of_colors * n_index + c_index + color)] = 2
            # Penalty for assigning the same color to adjacent nodes
            for node2 in list(nx.neighbors(problem_instance, node)):
                n2_index = nodes.index(node2)
                if n_index >= n2_index:
                    continue
                else:
                    QUBO[(amount_of_colors * n_index + c_index, amount_of_colors * n2_index + c_index)] = 1

    # If you want to see the qubo - uncomment the following lines
    # util.display_qubo(QUBO)

    return QUBO


'''
Note: this function works, as we always find valid solutions with QBSolv on small enough problems.
If the problem gets bigger or you try to display a solution received with the D-Wave - you will need to customize the
drawing function here
'''


def show_solution(solution_dict, output_path):
    """Save the solution (the final coloured graph) to a file.

    Parameters
    ----------
    solution_dict
        Dictionary containing the solution
    output_path
        Specifies the path to the file where the solution will be saved
    """
    print(solution_dict)
    # Creating the problem instance
    problem_instance = nx.Graph()
    problem_instance.add_edges_from([(0, 1), (0, 2), (1, 3), (2, 3)])
    color_labels = ["red", "green", "blue"]

    node_colors = []

    for colored_node in solution_dict:
        if solution_dict[colored_node] == 0:
            continue
        else:
            node_colors.append(color_labels[colored_node % len(color_labels)])

    nx.draw(problem_instance, with_labels=True, node_color=node_colors, font_color="white")
    print("-----------------\nThe solution file has been created: look at " + output_path + ".pdf\n-----------------")
    plt.savefig(output_path + ".pdf")



from dimod.binary_quadratic_model import BinaryQuadraticModel
import dimod
import dwave_networkx as dnx
import matplotlib.pyplot as plt


class Problem:
    """Class representing a problem in Ising or QUBO format. This class provides function for setting solving
    parameters and specific solvers and functions that belong to problems, e.g. find embeddings for this problem or
    solve the problem.

    Attributes
    ----------
    solver_params: dict
    uq_params: dict
        Dictionary that contains the key "num_repeats", which specifies the number of samplings the solver will process
    config
        Contains the users configuration data
    solver
        Specifies which solver should be used
    platform
        Specifies which platform should be used
    embedding
        Chimera or Pegasus embedding for the problem
    connection
        Connection object containing the configuration data of the user

    Methods
    -------
    with_solver(solver)
        Set the solver attribute to solver
    with_platform(platform)
        Set the platform attribute to platform
    with_params(**kw_params)
        Update the dictionary solver_params with the elements from kw_params
    with_uq_params(**kw_params)
        Update the dictionary uq_params with the elements from kw_params
    find_chimera_embedding()
        Call the connections find_chimera_embedding method to make a server request for finding a chimera
        embedding
    draw_chimera_embedding()
        Save the chimera embedding to a file.
    find_pegasus_embedding()
        Call the connections find_chimera_embedding method to make a server request for finding a pegasus
        embedding
    draw_pegasus_embedding()
        Save the pegasus embedding to a file.
    solve(times)
        Solve a problem by either calling the connections solve_qubo or solve_ising function.
    """

    def __init__(self, config):
        self.solver_params = {}
        self.uq_params = {}
        self.config = config
        self.solver = None
        self.platform = None
        self.embedding = None
        self.connection = config.create_connection()

    # ------------------ Set attributes ------------------ #

    def with_solver(self, solver):
        """Set the solver attribute"""
        self.solver = solver
        return self

    def with_platform(self, platform):
        """Set the platform attribute"""
        self.platform = platform
        return self

    def with_params(self, **kw_params):
        """Update the solver_params dictionary with the elements from kw_params"""
        self.solver_params.update(kw_params)
        return self

    def with_uq_params(self, **kw_params):
        """Update the uq_params dictionary with the elements from kw_params"""
        self.uq_params.update(kw_params)
        return self

    # ------------------ find embeddings ------------------ #

    def find_chimera_embedding(self):
        """Call the connections find_chimera_embedding method to make a server request for finding a chimera
        embedding """
        self.embedding = self.connection.find_chimera_embedding(self)
        return self.embedding

    def draw_chimera_embedding(self, output_path):
        """Save the chimera embedding to a file.

        Parameters
        ----------
        output_path
            Specifies the path to the file where to save the embedding
        """
        dnx.draw_chimera_embedding(dnx.chimera_graph(16, 16, 4), emb=self.embedding, node_size=3, width=.3)
        plt.savefig(output_path)

    def find_pegasus_embedding(self):
        """Call the connections find_chimera_embedding method to make a server request for finding a pegasus
        embedding """
        self.embedding = self.connection.find_pegasus_embedding(self)
        return self.embedding

    def draw_pegasus_embedding(self, output_path):
        """Save the pegasus embedding to a file.

        Parameters
        ----------
        output_path
            Specifies the path to the file where to save the embedding
        """
        dnx.draw_pegasus_embedding(dnx.pegasus_graph(11), emb=self.embedding, node_size=3, width=.3)
        plt.savefig(output_path)

    # ------------------ Solve problems ------------------ #

    def solve(self, times=1):
        """Solve a problem by either calling the connections solve_qubo or solve_ising function.

        Parameters
        ----------
        times: int
            Specifies the count of iterations. If no parameter is passed, the default value is 1.
        """

        self.uq_params.update({"num_repeats": times})
        if self.solver is not None:
            self.connection.set_preferred_solver(self.solver)
        if self.platform is not None:
            self.connection.set_preferred_platform(self.platform)

        if isinstance(self, Qubo):
            return self.connection.solve_qubo(self)
        if isinstance(self, Ising):
            return self.connection.solve_ising(self)


class Qubo(Problem):
    """Represents a problem in QUBO format.

    Attributes
    ----------
    problem_dict: dict
        QUBO represantation of a problem

    Methods
    -------
    to_json()
        Transform a QUBO dictionary into a dimod.BinaryQuadraticModel (BQM) and return the serialized BQM
    """
    def __init__(self, config, qubo_dict):
        Problem.__init__(self, config)
        self.problem_dict = qubo_dict

    def to_json(self):
        """Transform a QUBO dictionary into a dimod.BinaryQuadraticModel (BQM) and return the serialized BQM.

        Returns
        -------
        bqm.to_serializable(): dict
            The serialized BQM
        """
        linear = {}
        quadratic = {}
        for (a, b) in self.problem_dict.keys():
            if a == b:
                linear[a] = self.problem_dict[(a, b)]
            else:
                quadratic[(a, b)] = self.problem_dict[(a, b)]

        bqm = BinaryQuadraticModel(linear, quadratic, 0.0, dimod.BINARY)
        return bqm.to_serializable()


class Ising(Problem):
    """Represents a problem in Ising format.

    Attributes
    ----------
    linear_dict: dict
        external magnetic field values
    quadratic_dict: dict
        interaction values

    Methods
    -------
    to_json()
        Transform an Ising representation of a problem into a dimod.BinaryQuadraticModel (BQM) and return the serialized
        BQM
    """
    def __init__(self, config, linear_dict, quadratic_dict):
        Problem.__init__(self, config)
        self.linear_dict = linear_dict
        self.quadratic_dict = quadratic_dict

    def to_json(self):
        """
        Transform an Ising representation of a problem into a dimod.BinaryQuadraticModel (BQM) and return the serialized
        BQM.

        Returns
        -------
        bqm.to_serializable(): dict
            The serialized BQM
        """
        bqm = BinaryQuadraticModel(self.linear_dict, self.quadratic_dict, 0.0, dimod.SPIN)
        return bqm.to_serializable()



from .connection import Connection
import json


class Config:
    """A class that represents a config object containing the configuration data of the user. The config object is
    created at the beginning of execution (in main.py) either by a config file or by passed arguments specifying the
    configuration options.

    Attributes
    ----------
    endpoint
        endpoint (ip+port) to which the connection will be established
    method
        authentication method
    credentials
        personal token of the user

    The endpoint, method and credentials are specified in the config file or passed as a parameter to the config
    object in main.py.

    Methods
    ----------
    create_connection()
        Create a Connection object containing the configuration data of the user.
    """

    def __init__(self, **kwargs):
        """Create the config object and set the configuration data for the current user.

        Parameters
        ----------
        **kwargs
            **kwargs contains an unspecified number of arguments
            Either the arguments consist of the configuration data itself (endpoint, method, credentials)
            or the path to the config file containing the configuration data.
        """

        # authentication without config file
        if "endpoint" in kwargs:
            self.endpoint = kwargs["endpoint"]
        if "method" in kwargs and "credentials" in kwargs:
            self.method = kwargs["method"]
            self.credentials = kwargs["credentials"]
        elif "configpath" in kwargs:
            # authentication with config file
            with open(kwargs["configpath"]) as configfile:
                config = json.load(configfile)
                self.method = config["method"]
                self.credentials = config["credentials"]
                self.endpoint = config["endpoint"]

    def create_connection(self):
        """Create a connection object containing the configuration data of the user. """
        return Connection(self.endpoint, self.method, self.credentials)



from ..client.connection import Connection
from . import graph_coloring_example as graphcoloring
from .. import Problem

# Example QUBO with some reward and penalty values
example_qubo = {(0, 0): -2, (1, 1): -2, (2, 2): -2, (3, 3): -2, (4, 4): -2, (5, 5): -2,
                (0, 1): 5, (0, 2): 5, (0, 4): 5,
                (1, 3): 5, (1, 5): 5,
                (2, 3): 5, (2, 4): 5,
                (3, 5): 5,
                (4, 5): 5}

# Ising problem with some example values for external magnetic field (example_ising_h) and interaction (example_ising_J)
example_ising_h, example_ising_J = {1: 1, 2: 2, 3: 3}, {(1, 2): 4, (1, 3): 5, (2, 3): 6}


def ping(config):
    """Send a ping message to the server.

    Parameters
    ----------
    config
        The config object that contains the users configuration data
    """
    con = config.create_connection()
    print(con.ping())


def qbsolv_example_qubo(config):
    """Solve the QUBO example with QBSolv.

    Parameters
    ----------
    config
        The config object that contains the users configuration data
    """

    test_dict = {"test_parameter": {(0, 1): 5, (1, 2): 3}}
    answer = Problem.Qubo(config, example_qubo).with_platform("qbsolv").with_params(**test_dict).solve(100)

    # -----------
    # These calls return arrays containing the raw information in lists
    # -----------
    # answer.solutions
    # answer.energies
    # answer.num_occurrences

    # -----------
    # These functions will print the received information in different ways
    # -----------
    # answer.print_solutions()
    # answer.print_energies()
    # answer.print_num_occurrences()
    answer.print_solutions_nice()


def qbsolv_example_ising(config):
    """Solve the Ising example with QBSolv.

    Parameters
    ----------
    config
        The config object that contains the users configuration data
    """

    # Create the Ising representation of a problem we want to solve

    answer = Problem.Ising(config, example_ising_h, example_ising_J).with_platform("qbsolv").solve(48)

    # -----------
    # These calls return arrays containing the raw information in lists
    # -----------
    # answer.solutions
    # answer.energies
    # answer.num_occurrences

    # -----------
    # These functions will print the received information in different ways
    # -----------
    # answer.print_solutions()
    # answer.print_energies()
    # answer.print_num_occurrences()
    answer.print_solutions_nice()


def dwave_example_qubo(config):
    """Solve the QUBO example with the quantum solver 'DW_2000Q_6' from DWave.

    Parameters
    ----------
    config
        The config object that contains the users configuration data
    """

    answer = Problem.Qubo(config, example_qubo).with_platform("dwave").with_solver("DW_2000Q_6").solve(1)

    # -----------
    # These calls return arrays containing the raw information in lists
    # -----------
    # answer.solutions
    # answer.energies
    # answer.num_occurrences

    # -----------
    # These functions will print the received information in different ways
    # -----------
    # answer.print_solutions()
    # answer.print_energies()
    # answer.print_num_occurrences()
    answer.print_solutions_nice()


def dwave_example_ising(config):
    """Solve the Ising example with the quantum solver 'DW_2000Q_6' from DWave.

    Parameters
    ----------
    config
        The config object that contains the users configuration data
    """

    answer = Problem.Ising(config, example_ising_h, example_ising_J).with_platform("dwave") \
        .with_solver("DW_2000Q_6").solve(1)

    # -----------
    # These calls return arrays containing the raw information in lists
    # -----------
    # answer.solutions
    # answer.energies
    # answer.num_occurrences

    # -----------
    # These functions will print the received information in different ways
    # -----------
    # answer.print_solutions()
    # answer.print_energies()
    # answer.print_num_occurrences()

    answer.print_solutions_nice()


def get_dwave_solvers(config):
    """Show all available solvers from DWave. """
    connection = config.create_connection()
    print(connection.get_available_dwave_solvers())


def show_quota(config):
    """Print the users remaining quota. """
    connection = config.create_connection()
    connection.show_quota()


def graph_coloring_example(config):
    """Process an example of the graph coloring problem as a QUBO by using QBSolv. """
    QUBO = graphcoloring.create_qubo()
    answer = Problem.Qubo(config, QUBO).with_platform("qbsolv").solve(2)

    for index, solution in enumerate(answer.solutions):
        graphcoloring.show_solution(solution, "graph_coloring_solution_" + str(index))


def find_chimera_embedding_example(config):
    """Find a possible Chimera embedding by calling the DW_2000Q_6 solver from DWaveand save it to embedding.pdf. """
    problem = Problem.Qubo(config, example_qubo).with_platform("dwave").with_solver("DW_2000Q_6")
    # Note: If you generate an embedding via problem.find_chimera_embedding(), the embedding will be saved in the
    #       variable problem.embedding - you can directly access, view and save it.
    problem.find_chimera_embedding()
    problem.draw_chimera_embedding("embedding.pdf")


def find_pegasus_embedding_example(config):
    """Find a possible Pegasus embedding by calling the Advantage_system1.1 solver from DWave and save it to
    embedding.pdf. """
    problem = Problem.Qubo(config, example_qubo).with_platform("dwave").with_solver("Advantage_system1.1")
    # Note: If you generate an embedding via problem.find_chimera_embedding(), the embedding will be saved in the
    #       variable problem.embedding - you can directly access, view and save it.
    problem.find_pegasus_embedding()
    problem.draw_pegasus_embedding("embedding_pegasus.pdf")


def dwave_example_qubo_with_custom_embedding(config):
    """Solve an example QUBO problem with a given embedding. """

    problem = Problem.Qubo(config, example_qubo).with_platform("dwave").with_solver("Advantage_system1.1")

    # Note: If you generate an embedding via problem.find_chimera_embedding(), the embedding will be saved in the
    #       variable problem.embedding - you can directly access, view and save it.
    problem.find_chimera_embedding()

    # This will automatically use the embedding generated with problem.find_chimera_embedding()
    # In general: if you have an embedding already - just assign this embedding to the problem.embedding variable
    # If you call problem.solve after, UQ will automatically use this embedding to solve the problem on the D-Wave
    answer = problem.solve(2)

    # -----------
    # These calls return arrays containing the raw information in lists
    # -----------
    # answer.solutions
    # answer.energies
    # answer.num_occurrences

    # -----------
    # These functions will print the received information in different ways
    # -----------
    # answer.print_solutions()
    # answer.print_energies()
    # answer.print_num_occurrences()

    answer.print_solutions_nice()



from dimod.sampleset import SampleSet
from prettytable import PrettyTable


class Response:
    """A Response object is created for a response to a solving task (that is solved by QBSolv or a DWave solver).
    It contains the sampleset (solution vectors, energy and number of occurrences), a list of the sampled solution
    vectors (solutions), the energies and the number of occurrences of each solution vector.

    Attributes
    ----------
    sampleset
        Table with solution vectors, energy and number of occurrences
    solutions
        List of the sampled solution vectors
    energies
        List of the solution vectors energies
    num_occurrences
        List of number of occurrences of a solution vector

    Methods
    -------
    print_solutions(), print_energies(), print_num_occurrences()
        Print the solution vectors, the energies and the number of occurrences of the vectors.
    print_solutions_nice()
        Show the solution (solution vectors, energies and number of occurrences) in a well readable table format.
    """

    def __init__(self, sampleset):
        self.sampleset = sampleset
        self.solutions = list(self.sampleset.samples())
        self.energies = list(map(lambda x: x.energy, list(self.sampleset.data(fields=["energy"]))))
        self.num_occurrences = list(
            map(lambda x: x.num_occurrences, list(self.sampleset.data(fields=["num_occurrences"]))))

    def print_solutions(self):
        for solution in self.solutions:
            print(solution)

    def print_energies(self):
        for energy in self.energies:
            print(energy)

    def print_num_occurrences(self):
        for occurrence in self.num_occurrences:
            print(occurrence)

    def print_solutions_nice(self):
        """Show the solution (solution vectors, energies and number of occurrences) in a well readable table format. """
        t = PrettyTable(["Answer-Sample", "Energy", "Num-Occurrences"])

        for index, solution in enumerate(self.solutions):
            t.add_row([solution, self.energies[index], self.num_occurrences[index]])

        print(t)


class QBSolveResponse(Response):
    """Response that represents a reply from the QBSolv Solver. """

    def __init__(self, dimod_answer):
        solution = SampleSet.from_serializable(dimod_answer)
        Response.__init__(self, solution)


class DWaveResponse(Response):
    """Response that represents a reply from the DWave Solver. """

    def __init__(self, dwave_answer):
        sampleset = SampleSet.from_serializable(dwave_answer)
        Response.__init__(self, sampleset)
        self.timing = self.sampleset.info["timing"]



class UQOException(Exception):
    def __init__(self, message):
        Exception.__init__(self, "\n\n" + "-"*10 + message + "\n\n" + "-"*10)


# ------------ CLIENT - EXCEPTIONS ------------ #


class NotAQuboException(UQOException):
    def __init__(self):
        message = "\n\nThe specified Problem is not a uq.Problem.Qubo instance"
        UQOException.__init__(self, message)


# ------------ AUTH - EXCEPTIONS ------------ #


class MissingTaskException(UQOException):
    def __init__(self):
        message = "\n\nNo task specified"
        UQOException.__init__(self, message)


class InvalidTaskException(UQOException):
    def __init__(self, error_details):
        message = "\n\nInvalid task: '" + str(error_details["parameters_sent"]) + \
                  "'\nAvailable tasks are: " + str(list(map(lambda x: str(x), error_details["tasks_available"])))
        UQOException.__init__(self, message)


class InvalidAuthMethodException(UQOException):
    def __init__(self, error_details):
        message = "\n\nAuthentication failed\nInvalid authentication method: '" + str(error_details["parameters_sent"]) + \
                  "'\nAvailable authentication methods are: " + str(list(map(lambda x: str(x), error_details["auth_methods_available"])))
        UQOException.__init__(self, message)


class InvalidAuthCredentials(UQOException):
    def __init__(self):
        message = "\n\nAuthentication failed\nInvalid credentials"
        UQOException.__init__(self, message)


class AuthMessageMissingException(UQOException):
    def __init__(self):
        message = "\n\nAuthentication failed\nNo authentication information provided"
        UQOException.__init__(self, message)


class MalformedAuthMessageException(UQOException):
    def __init__(self):
        message = "\n\nAuthentication failed\nMalformed authentication message"
        UQOException.__init__(self, message)


class FastRetryException(UQOException):
    def __init__(self,answer_details):
        message = "\n\nYou sent too many QUBOs.\nYou are allowed to solve one QUBO every %d seconds" %(answer_details["interval"])
        UQOException.__init__(self, message)


class AuthAdminFailedException(UQOException):
    def __init__(self):
        message = "\n\nInvalid user or insufficient permissions"
        UQOException.__init__(self, message)

class GenericAuthException(UQOException):
    def __init__(self):
        message = "\n\nAn unexpected error occured\nPlease try again later!\n\nShould this problem still occur in some hours please contact the developer: sebastian.zielinski@ifi.lmu.de"
        UQOException.__init__(self, message)


# ------------  BACKEND - SERVER ------------ #

class GenericBackendException(UQOException):
    def __init__(self):
        message = "\n\nAn unexpected error occured\nPlease try again later"
        UQOException.__init__(self, message)


class SolverException(UQOException):
    def __init__(self, answer_details):
        message = "\n\nAn error occured while solving the problem\n\n"+answer_details["message"]
        UQOException.__init__(self, message)


class MissingTask(UQOException):
    def __init__(self):
        message = "\n\nNo task was specified"
        UQOException.__init__(self, message)


class InvalidTask(UQOException):
    def __init__(self, answer_details):
        message = "\n\nAn invalid task was submitted.\nSubmitted task: " + answer_details["value"]
        UQOException.__init__(self, message)


class QBSolveException(UQOException):
    def __init__(self, message):
        UQOException.__init__(self, message)


class InsufficientQuotaException(UQOException):
    def __init__(self):
        message = "\n\nYour quota is insufficient to perform this task."
        UQOException.__init__(self, message)


class InvalidSolverException(UQOException):
    def __init__(self, error_details):
        message = "\n\nThere is no solver named '" + error_details["parameters_sent"] + "' for the platform '" + error_details["platform"] +"'"
        UQOException.__init__(self, message)


class MissingPlatformException(UQOException):
    def __init__(self, error_details):
        message ="\n\nNo platform was specified.\n" \
                 "Valid platforms are: "
        for platform in error_details["available_platforms"]:
            message += "'" + platform + "'"
        UQOException.__init__(self, message)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="uqo",
    version="0.0.2.1",
    author="QAR-Lab Munich",
    author_email="Sebastian.Zielinski@ifi.lmu.de",
    description="Client of the optimization framework UQO",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    packages=setuptools.find_packages(),
    install_requires=[
        'dimod==0.9.13',
        'dwave-networkx==0.8.8',
        'prettytable==2.0.0',
        'pyzmq==22.0.2',
        'matplotlib==3.3.2'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)